from dataclasses import field, dataclass
from typing import Union, Sized, Dict, List, Tuple, Iterable

import pandas as pd

from powerplantsim import utils
from powerplantsim.datatypes import Machine, Storage, Customer, Purchaser, Supplier, Edge, Node
from powerplantsim.utils.typing import NodeID, Plan, StepPlan, State, Flow, EdgeID, NamedTuple

States = Dict[Machine, Union[State, pd.Series]]
"""Datatype for states."""

Flows = Dict[Edge, Union[Flow, pd.Series]]
"""Datatype for flows."""


@dataclass(frozen=True, unsafe_hash=True, slots=True)
class StepPlanInfo(NamedTuple):
    """Datatype for the definition of the plan for a single step."""

    states: Dict[Machine, State] = field(default_factory=dict, init=False)
    """The states for the given step."""

    flows: Dict[Edge, Flow] = field(default_factory=dict, init=False)
    """The flows for the given step."""


class SimulationOutput:
    """Dataclass representing the output of a power plant simulation."""

    def __init__(self, horizon: pd.Index):
        """
        :param horizon:
            The time horizon of the simulation.
        """
        self.flows: pd.DataFrame = pd.DataFrame(index=horizon, dtype=float)
        """The commodities flows indexed by edge (source, destination)."""

        self.states: pd.DataFrame = pd.DataFrame(index=horizon, dtype=float)
        """The machine setpoints indexed by machine name."""

        self.storage: pd.DataFrame = pd.DataFrame(index=horizon, dtype=float)
        """The commodities that were stored in the plant indexed by storage name."""

        self.demands: pd.DataFrame = pd.DataFrame(index=horizon, dtype=float)
        """The true demands indexed by customer name."""

        self.buying_prices: pd.DataFrame = pd.DataFrame(index=horizon, dtype=float)
        """The true buying prices indexed by purchaser name."""

        self.sell_prices: pd.DataFrame = pd.DataFrame(index=horizon, dtype=float)
        """The true selling prices indexed by supplier name."""


def check_plant(plant):
    """Checks that the plant is valid, i.e., that each commodity flows correctly and has a start and end point.
    If the plant is not valid, raises an exception.

    :param plant:
        The plant to check.
    """
    for name, node in plant.nodes().items():
        for commodity in node.commodities_in:
            edges = plant.edges(destinations=name, commodities=commodity)
            assert len(edges) > 0, f"Input commodity {commodity} has no valid ingoing edge in node {name}"
        for commodity in node.commodities_out:
            edges = plant.edges(sources=name, commodities=commodity)
            assert len(edges) > 0, f"Output commodity {commodity} has no valid outgoing edge in node {name}"


# noinspection PyTypeChecker
def process_plan(plan: Union[Plan, pd.DataFrame],
                 machines: Dict[NodeID, Machine],
                 edges: Dict[EdgeID, Edge],
                 horizon: pd.Index) -> Tuple[List[StepPlanInfo], Dict[NodeID, pd.Series], Dict[EdgeID, pd.Series]]:
    """Checks that the given plan is correctly specified and converts it to a standard format.

    :param plan:
        The energetic plan of the power plant defined as vectors of states/flows.

    :param machines:
        The dictionary of all the machine nodes in the plant for which a vector of states must be provided.

    :param edges:
        The dictionary of all the edges in the plant for which a vector of flows must be provided.

    :param horizon:
        The time horizon of the simulation.

    :return:
        A tuple where the first element is the energetic plan as a list of step plans, while the remaining two elements
        are dictionaries of states/flows indexed by nodes/edges.
    """
    # convert dictionary to dataframe if needed, and check consistency of flow vectors
    if isinstance(plan, dict):
        df = pd.DataFrame(index=horizon)
        for datatype, vector in plan.items():
            assert not isinstance(vector, Sized) or len(vector) == len(horizon), \
                f"Vector for key '{datatype}' has length {len(vector)}, expected {len(horizon)}"
            df[datatype] = pd.Series(vector, index=horizon)
        plan = df
    # distinguish between states and flows while checking that all the datatypes (columns) are in the given sets
    # and return a concatenated version of the states and flows dataframes with a higher level column index
    states, flows = check_plan(plan=plan, machines=machines, edges=edges)
    output = [StepPlanInfo() for _ in range(len(plan))]
    for machine, values in states.items():
        for i, value in enumerate(values):
            output[i].states[machine] = value
    for edge, values in flows.items():
        for i, value in enumerate(values):
            output[i].flows[edge] = value
    return output, states, flows


def check_plan(plan: Union[StepPlan, pd.Series, Plan, pd.DataFrame],
               machines: Dict[NodeID, Machine],
               edges: Dict[EdgeID, Edge]) -> Tuple[States, Flows]:
    """Checks that the given plan is correctly specified and returns the dictionaries of states and flows.

    :param plan:
        The energetic plan of the power plant defined as vectors of states/flows.

    :param machines:
        The dictionary of all the machine nodes in the plant for which a vector of states must be provided.

    :param edges:
        The dictionary of all the edges in the plant for which a vector of flows must be provided.

    :return:
        A tuple (states, flows) containing the dictionaries of states/flows indexed by machine/edge.
    """
    machines, edges = machines.copy(), edges.copy()
    states, flows = {}, {}
    for key, value in plan.items():
        if key in machines:
            mac = machines.pop(key)
            states[mac] = value
        elif key in edges:
            edge = edges.pop(key)
            flows[edge] = value
        else:
            raise AssertionError(f"Key {utils.stringify(key)} is not present in the plant")
    # check that the remaining sets are empty, i.e., there is no missing states/flows in the dataframe
    assert len(machines) == 0, f"No states vector has been passed for machines {list(machines.keys())}"
    assert len(edges) == 0, f"No flows vector has been passed for edges {list(edges.keys())}"
    return states, flows


def build_output(nodes: Iterable[Node], edges: Iterable[Edge], horizon: pd.Index) -> SimulationOutput:
    """Builds the output of the simulation.

    :param nodes:
        The simulation nodes.

    :param edges:
        The simulation edges.

    :param horizon:
        The time horizon of the simulation.

    :return:
        A SimulationOutput object containing all the information about true prices, demands, setpoints, and storage.
    """
    output = SimulationOutput(horizon=horizon)
    for edge in edges:
        output.flows[edge.key] = edge.flows
    for node in nodes:
        if isinstance(node, Machine):
            output.states[node.name] = node.states
        elif isinstance(node, Storage):
            output.storage[node.name] = node.storage
        elif isinstance(node, Customer):
            output.demands[node.name] = node.values
        elif isinstance(node, Purchaser):
            output.buying_prices[node.name] = node.values
        elif isinstance(node, Supplier):
            output.sell_prices[node.name] = node.values
        else:
            raise AssertionError(f"Unknown node type {type(node)}")
    return output



from dataclasses import dataclass, field
from typing import Dict, Iterable, List, Union, Any, Optional

import matplotlib.pyplot as plt
import networkx as nx
from matplotlib.lines import Line2D

from powerplantsim import utils
from powerplantsim.utils import NamedTuple
from powerplantsim.utils.typing import SingleEdgeID


@dataclass(frozen=True, unsafe_hash=True, slots=True, kw_only=True)
class StyleInfo(NamedTuple):
    """Datatype for nodes/edges style information."""

    color: str = field(kw_only=True)
    """The color of the node/edge."""

    shape: str = field(kw_only=True)
    """The shape of the node/edge."""


NODE_STYLES: Dict[str, StyleInfo] = {
    'supplier': StyleInfo(color='#FFEAB8', shape='>'),
    'customer': StyleInfo(color='#A6DEAE', shape='<'),
    'purchaser': StyleInfo(color='#FFEAB8', shape='<'),
    'machine': StyleInfo(color='#FF700A', shape='o'),
    'storage': StyleInfo(color='#ABA5E8', shape='s')
}
"""Default dictionary of nodes style information."""


def get_node_positions(graph: nx.DiGraph,
                       sources: Iterable[str],
                       node_pos: Union[str, List[Iterable[Optional[str]]], Dict[str, Any]]) -> dict:
    """Traverse the graph from the sources and label each node with a progressive number the position of the nodes is
    eventually obtained by layering them respectively to the computed number so that sources will be on the left and
    sinks on the right.

    :param graph:
        The networkx DiGraph instance.

    :param sources:
        The iterable of source nodes.

    :param node_pos:
        Information about the node position.

    :return:
        A dictionary of node positions.
    """
    # if the position info is a dictionary, it already contains the mapping <node: pos> so return the info itself
    if isinstance(node_pos, dict):
        return node_pos
    # if the position info is a list, create a new graph with nodes added sequentially and with the layers attribute
    if isinstance(node_pos, list):
        p = 0
        nodes = set(graph.nodes)
        edges = set(graph.edges)
        graph = nx.DiGraph()
        for layer, nodelist in enumerate(node_pos):
            for node in nodelist:
                if node is None:
                    p += 1
                    node = f'__placeholder-{p}__'
                else:
                    assert node in nodes, f"Node {node} is not present in the given plant"
                graph.add_node(node_for_adding=node, layer=layer)
        graph.add_edges_from(edges)
        pos = nx.multipartite_layout(graph, subset_key='layer')
        return {k: v for k, v in pos.items() if k in nodes}
    # if the position info is a string representing the layering strategy, add the layers to each node accordingly
    graph = graph.copy()
    if node_pos == 'sp':
        # use breadth first search for shortest paths
        for it, nodes in enumerate(nx.bfs_layers(graph, sources=sources)):
            for node in nodes:
                graph.nodes[node]['layer'] = it
    elif node_pos == 'lp':
        # use floyd warshall algorithm to search for longest paths
        #  - get the indices of the sources
        #  - get the negative shortest path matrix and select only the paths from the sources
        #  - get the negative minimum value for each node in the plant and negate it to get the layer
        nx.set_edge_attributes(graph, values=-1, name='weight')
        sources = set(sources)
        sources = [i for i, node in enumerate(graph.nodes) if node in sources]
        lp = -nx.floyd_warshall_numpy(graph)[sources].min(axis=0)
        for i, node in enumerate(graph.nodes):
            graph.nodes[node]['layer'] = lp[i]
    else:
        raise AssertionError(f"Unsupported node_pos: {node_pos}")
    return nx.multipartite_layout(graph, subset_key='layer')


def get_node_style(colors: Union[None, str, Dict[str, str]],
                   markers: Union[None, str, Dict[str, str]]) -> Dict[str, StyleInfo]:
    """Builds a dictionary of color and marker mappings indexed by node kind.
    In case either the colors or the shapes are not None, include the custom information in the dictionary.

    :param colors:
        Either a string representing the color of the nodes, or a dictionary {kind: color} which associates a color to
        each node kind ('supplier', 'client', 'machine').

    :param markers:
        Either a string representing the shape of the nodes, or a dictionary {kind: shape} which associates a shape to
        each node kind ('supplier', 'client', 'machine', 'storage').

    :return:
        The dictionary of style information.
    """
    styles = {}
    for kind, style in NODE_STYLES.items():
        color = utils.get_matching_object(matcher=colors, index=kind, default=style.color)
        marker = utils.get_matching_object(matcher=markers, index=kind, default=style.shape)
        styles[kind] = StyleInfo(color=color, shape=marker)
    return styles


def get_edge_style(colors: Union[None, str, Dict[str, str]],
                   shapes: Union[None, str, Dict[str, str]],
                   commodities: List[str]) -> Dict[str, StyleInfo]:
    """Build a dictionary of color and style mappings indexed by commodity.

    :param colors:
        Either a string representing the color of the commodity, or a dictionary {kind: color}.

    :param shapes:
        Either a string representing the shape of the commodity, or a dictionary {kind: shape}.

    :param commodities:
        The list of all the commodities in the plant.

    :return:
        The dictionary of style information.
    """
    styles = {}
    for com in commodities:
        color = utils.get_matching_object(matcher=colors, index=com, default='black')
        marker = utils.get_matching_object(matcher=shapes, index=com, default='solid')
        styles[com] = StyleInfo(color=color, shape=marker)
    return styles


def build_node_label(kind: str, style: StyleInfo) -> Line2D:
    """Create a label for the legend of nodes with the correct style, color, and text.

    :param kind:
        The type of nodes.

    :param style:
        The style information.

    :return:
        A Line2D object representing the label.
    """
    return Line2D(
        [],
        [],
        marker=style.shape,
        markerfacecolor=style.color,
        markeredgecolor='black',
        linestyle='None',
        markersize=25,
        label=kind.title()
    )


def build_edge_label(commodity: str, style: StyleInfo) -> Line2D:
    """Create a label for the legend of edges with the correct style, color, and text.

    :param commodity:
        The type of commodity flowing in the edge.

    :param style:
        The style information.

    :return:
        A Line2D object representing the label.
    """
    return Line2D(
        [],
        [],
        lw=2,
        color=style.color,
        linestyle=style.shape,
        label=commodity.title()
    )


def draw_nodes(graph: nx.DiGraph,
               pos: dict,
               nodes: Iterable[str],
               style: StyleInfo,
               size: float,
               width: float,
               ax: plt.Axes):
    """Draws a subset of nodes from the plant.

    :param graph:
        The networkx DiGraph instance.

    :param pos:
        The dictionary of node's positions.

    :param nodes:
        The subset of nodes to be drawn.

    :param style:
        The style information.

    :param size:
        The size of the nodes.

    :param width:
        The width of the node's borders.

    :param ax:
        The ax on which to plot.
    """
    nx.draw(
        graph,
        pos=pos,
        edgelist=[],
        nodelist=nodes,
        node_color=style.color,
        node_shape=style.shape,
        node_size=size * 100,
        linewidths=width,
        with_labels=True,
        edgecolors='k',
        arrows=True,
        ax=ax
    )


def draw_edges(graph: nx.DiGraph,
               pos: dict,
               edges: Iterable[SingleEdgeID],
               style: StyleInfo,
               size: float,
               width: float,
               ax: plt.Axes):
    """Draws a subset of edges from the plant.

    :param graph:
        The networkx DiGraph instance.

    :param pos:
        The dictionary of node's positions.

    :param edges:
        The subset of edges to be drawn.

    :param style:
        The style information.

    :param size:
        The size of the nodes.

    :param width:
        The width of the edges.

    :param ax:
        The ax on which to plot.
    """
    nx.draw(
        graph,
        pos=pos,
        nodelist=[],
        edgelist=edges,
        edge_color=style.color,
        style=style.shape,
        node_size=size * 100,
        arrowsize=width * 10,
        width=width,
        arrows=True,
        ax=ax
    )



from dataclasses import dataclass, field
from typing import Set, List, Optional, Dict, Any

import numpy as np
import pandas as pd
import pyomo.environ as pyo
# noinspection PyPackageRequirements
from descriptors import classproperty

from powerplantsim.datatypes.node import Node
from powerplantsim.utils.typing import Flow, State


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class Storage(Node):
    """A node in the plant that stores certain commodities."""

    commodity: str = field(kw_only=True)
    """The commodity stored by the machine."""

    capacity: float = field(kw_only=True)
    """The storage capacity, which must be a strictly positive and finite number."""

    dissipation: float = field(kw_only=True)
    """The dissipation rate of the storage at every time unit, which must be a float in [0, 1]."""

    charge_rate: float = field(kw_only=True)
    """The maximal charge rate (input flow) in a time unit."""

    discharge_rate: float = field(kw_only=True)
    """The maximal discharge rate (output flow) in a time unit."""

    _storage: List[float] = field(init=False, default_factory=list)
    """The series of actual commodities storage, which is filled during the simulation."""

    def __post_init__(self):
        self._info['current_storage'] = None
        assert self.capacity != float('inf'), "Capacity should be a finite number, got inf"
        assert self.capacity > 0.0, f"Capacity should be strictly positive, got {self.capacity}"
        assert self.charge_rate > 0.0, f"Charge rate should be strictly positive, got {self.charge_rate}"
        assert self.discharge_rate > 0.0, f"Discharge rate should be strictly positive, got {self.discharge_rate}"
        assert 0.0 <= self.dissipation <= 1.0, f"Dissipation should be in range [0, 1], got {self.dissipation}"

    @classproperty
    def kind(self) -> str:
        return 'storage'

    @classproperty
    def _properties(self) -> List[str]:
        properties = super(Storage, self)._properties
        return properties + [
            'commodity',
            'capacity',
            'dissipation',
            'charge_rate',
            'discharge_rate',
            'storage',
            'current_storage'
        ]

    @property
    def storage(self) -> pd.Series:
        """The series of actual commodities storage, which is filled during the simulation."""
        return pd.Series(self._storage, dtype=float, index=self._horizon[:len(self._storage)])

    @property
    def current_storage(self) -> Optional[float]:
        """The current storage of the node for this time step."""
        return self._info['current_storage']

    @property
    def commodities_in(self) -> Set[str]:
        return {self.commodity}

    @property
    def commodities_out(self) -> Set[str]:
        return {self.commodity}

    # noinspection PyTypeChecker
    def to_pyomo(self, mutable: bool = False) -> pyo.Block:
        # start from the default node block and create aliases for the flows of the (unique) commodity
        node = super(Storage, self).to_pyomo(mutable=mutable)
        in_flow, out_flow = node.in_flows[self.commodity], node.out_flows[self.commodity]
        # add a parameter representing the current storage (and initialize it if needed)
        kwargs = dict(mutable=True) if mutable else dict(initialize=self.current_storage)
        node.current_storage = pyo.Param(domain=pyo.NonNegativeReals, **kwargs)
        # model the storage as the sum of between the current storage and the input and output flows
        node.storage = node.current_storage + in_flow - out_flow
        node.capacity_lb = pyo.Constraint(rule=node.storage >= 0)
        node.capacity_ub = pyo.Constraint(rule=node.storage <= self.capacity)
        # impose constraints on either input or output flows
        #  - create a binary variable where 0 means that there is an output flow, 1 means that there is an input flow
        #  - impose big-M constraints on input and output flows using the change/discharge rates as M
        #  - charge/discharge rates are lower bounded by remaining storage (capacity - current storage) and current
        #    storage (or total capacity if not available) to account for the fact that they can have infinite values
        #  - this definition automatically allow to impose the charge/discharge rate upper bounds
        charge_rate = min(self.capacity - (0 if mutable else self.current_storage), self.charge_rate)
        discharge_rate = min(self.capacity if mutable else self.current_storage, self.discharge_rate)
        node.flow_selector = pyo.Var(domain=pyo.Binary, initialize=0)
        node.input_flow_cst = pyo.Constraint(rule=in_flow <= node.flow_selector * charge_rate)
        node.output_flow_cst = pyo.Constraint(rule=out_flow <= (1 - node.flow_selector) * discharge_rate)
        return node

    def update(self, rng: np.random.Generator, flows: Dict[Any, Flow], states: Dict[Any, State]):
        self._info['current_storage'] = 0.0 if len(self._storage) == 0 else (1 - self.dissipation) * self._storage[-1]

    def step(self, flows: Dict[Any, Flow], states: Dict[Any, State]):
        # compute total input and output flows from respective edges
        in_flow, out_flow = 0.0, 0.0
        for edge, flow in flows.items():
            in_flow += flow if edge.destination == self.name else 0.0
            out_flow += flow if edge.source == self.name else 0.0
        # check that at least one of the two is null as from the constraints
        assert in_flow == 0.0 or out_flow == 0.0, \
            f"Storage node '{self.name}' can have either input or output flows in a single time step, got both"
        # check that the input and output flows are compatible with the charge rates
        assert in_flow <= self.charge_rate + self.eps, \
            f"Storage node '{self.name}' should have maximal input flow {self.charge_rate}, got {in_flow}"
        assert out_flow <= self.discharge_rate + self.eps, \
            f"Storage node '{self.name}' should have maximal output flow {self.discharge_rate}, got {out_flow}"
        # compute and check new storage from previous one (discounted by 1 - dissipation) and difference between flows
        storage = self._info['current_storage'] + in_flow - out_flow
        assert storage >= -self.eps, f"Storage node '{self.name}' cannot contain negative amount, got {storage}"
        assert storage <= self.capacity + self.eps, \
            f"Storage node '{self.name}' cannot contain more than {self.capacity} amount, got {storage}"
        self._storage.append(np.clip(storage, a_min=0, a_max=self.capacity))
        self._info['current_storage'] = None



from dataclasses import dataclass, field
from typing import Set, Optional, Tuple, List, Dict, Any

import numpy as np
import pandas as pd
import pyomo.environ as pyo
# noinspection PyPackageRequirements
from descriptors import classproperty
from pyomo.core import Piecewise

from powerplantsim import utils
from powerplantsim.datatypes.node import Node
from powerplantsim.utils.typing import State, Flow


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class Machine(Node):
    """A node in the plant that converts certain commodities in others."""

    _setpoint: pd.DataFrame = field(kw_only=True)
    """A pandas dataframe where the index is a series of floating point values that indicate the input commodity flow,
    while the columns should be named after the output commodity and contain the respective output flows."""

    discrete_setpoint: bool = field(kw_only=True)
    """Whether the setpoint is discrete or continuous."""

    max_starting: Optional[Tuple[int, int]] = field(kw_only=True)
    """A tuple <N, T> where N is the maximal number of times that the machine can be switched on in T units."""

    cost: float = field(kw_only=True)
    """The cost for operating the machine (cost is discarded when the machine is off)."""

    _states: List[State] = field(init=False, default_factory=list)
    """The series of actual input setpoints (None for machine off), which is filled during the simulation."""

    def __post_init__(self):
        self._info['current_state'] = None
        # sort setpoint and rename index
        self._setpoint.sort_index(inplace=True)
        self._setpoint.index.rename(name='setpoint', inplace=True)
        # check that set points are strictly positive (the first one is enough since it is sorted)
        assert self._setpoint.index[0] >= 0.0, f"Setpoints should be non-negative, got {self._setpoint.index[0]}"
        for c in self._setpoint.columns:
            # check that the column is structured as a tuple ('input'|'output', commodity)
            assert isinstance(c, tuple) and len(c) == 2 and c[0] in ['input', 'output'], \
                f"Setpoint columns should be a tuple ('input'|'output', commodity), got {utils.stringify(c)}"
            # check that the corresponding output flows are non-negative (take the minimum value for each column)
            lb = self._setpoint[c].min()
            assert lb >= 0.0, f"Setpoint flows should be non-negative, got {c}: {lb}"
        # check max starting
        if self.max_starting is not None:
            n, t = self.max_starting
            assert n > 0, f"The number of starting must be a strictly positive integer, got {n}"
            assert n < t, f"The number of starting must be strictly less than the number of time steps, got {n} >= {t}"
        # check non-negative cost
        assert self.cost >= 0.0, f"The operating cost of the machine must be non-negative, got {self.cost}"

    @classproperty
    def kind(self) -> str:
        return 'machine'

    @classproperty
    def _properties(self) -> List[str]:
        properties = super(Machine, self)._properties
        return properties + [
            'commodities_in',
            'commodities_out',
            'setpoint',
            'discrete_setpoint',
            'max_starting',
            'cost',
            'states',
            'current_state'
        ]

    def starts(self, t: int) -> int:
        """Computes the number of times the machine has been started in the past <t> steps.

        :param t:
            The number of past steps to consider.

        :return:
            The number of machine starts.
        """
        if t <= 0:
            return 0
        count = 0
        # create a list of the last T states
        t = min(t, len(self._states))
        # prepend nan (machine starts off) and append the last one
        states = [np.nan, *self._states[-t:]]
        # check consecutive pairs and increase the counter if we pass from a NaN to a real number
        for s1, s2 in zip(states[:-1], states[1:]):
            if np.isnan(s1) and not np.isnan(s2):
                count += 1
        return count

    @property
    def states(self) -> pd.Series:
        """The series of actual input setpoints (NaN for machine off), which is filled during the simulation."""
        return pd.Series(self._states, dtype=float, index=self._horizon[:len(self._states)])

    @property
    def current_state(self) -> Optional[State]:
        """The current state of the machine for this time step as provided by the user."""
        return self._info['current_state']

    @property
    def previous_state(self) -> State:
        """The state of the machine in the previous time step (or np.nan if this is the first time step)."""
        return np.nan if len(self._states) == 0 else self._states[-1]

    @property
    def setpoint(self) -> pd.DataFrame:
        """A pandas dataframe where the index is a series of floating point values that indicate the input commodity
        flow, while the columns should be named after the output commodity and contain the respective output flows."""
        return self._setpoint.copy()

    @property
    def commodities_in(self) -> Set[str]:
        return set(self._setpoint['input'].columns)

    @property
    def commodities_out(self) -> Set[str]:
        return set(self._setpoint['output'].columns)

    # noinspection PyTypeChecker
    def to_pyomo(self, mutable: bool = False) -> pyo.Block:
        # start from the default node block and store aliases for setpoint lower and upper bounds
        node = super(Machine, self).to_pyomo(mutable=mutable)
        lb, ub = self._setpoint.index[[0, -1]]
        # add a parameter representing the current state (and initialize it if needed)
        current_state = self.current_state
        kwargs = dict(mutable=True) if mutable else dict(initialize=lb if np.isnan(current_state) else current_state)
        node.current_state = pyo.Param(domain=pyo.NonNegativeReals, **kwargs)
        # create a variable for the machine binary state (on/off) then, if the number of starts in the last <t - 1>
        # steps is already maximal, set the variable bounds as (0, 0) so that the machine will be forced to be off
        n, t = (1, 1) if self.max_starting is None else self.max_starting
        node.switch = pyo.Var(domain=pyo.Binary, bounds=(0, 1 if self.starts(t=t - 1) < n else 0), initialize=0)
        # compute the cost of operating the machine (in case it is on)
        node.cost = self.cost * node.switch
        # handle setpoint for discrete and continuous machines + trivial case with a unique setpoint
        kwargs = dict() if mutable else dict(initialize=lb if np.isnan(current_state) else current_state)
        if len(self._setpoint) <= 1:
            # the state is 0 if the machine is off, otherwise it is the unique state, same for the flows
            node.state = pyo.Var(domain=pyo.NonNegativeReals, bounds=(0, ub), **kwargs)
            node.state_cst = pyo.Constraint(rule=node.state == node.switch * self._setpoint.index[0])
            node.input_flows_cst = pyo.Constraint(
                node.in_flows.index_set(),
                rule=lambda _, com: node.in_flows[com] == node.switch * self._setpoint.input[com].iloc[0]
            )
            node.output_flows_cst = pyo.Constraint(
                node.out_flows.index_set(),
                rule=lambda _, com: node.out_flows[com] == node.switch * self._setpoint.output[com].iloc[0]
            )
        elif self.discrete_setpoint:
            # build a one-hot encoded selector that has a single entry if the machine is on (i.e., node.switch == 1)
            # or no entry if the machine is off (i.e., node.switch == 0)
            node.selector = pyo.Var(range(len(self._setpoint)), domain=pyo.Binary, initialize=0)
            node.selector_cst = pyo.Constraint(rule=sum(node.selector.values()) == node.switch)
            # build a variable for the actual setpoint so that it is equal to the value indexed by the selector
            #  - use a variable instead of a plain equation in order to access it via the ".value" property
            node.state = pyo.Var(domain=pyo.NonNegativeReals, bounds=(0, ub), **kwargs)
            node.state_cst = pyo.Constraint(rule=node.state == sum(node.selector * self._setpoint.index))
            # impose constraints on input/output flows so that they match the correct setpoint indexed by the selector
            node.input_flows_cst = pyo.Constraint(
                node.in_flows.index_set(),
                rule=lambda _, com: node.in_flows[com] == sum(node.selector * self._setpoint.input[com])
            )
            node.output_flows_cst = pyo.Constraint(
                node.out_flows.index_set(),
                rule=lambda _, com: node.out_flows[com] == sum(node.selector * self._setpoint.output[com])
            )
        else:
            # build a state variable that is bounded within the min and max setpoint
            breakpoints = self._setpoint.index.values
            node.state = pyo.Var(domain=pyo.NonNegativeReals, bounds=(lb, ub), **kwargs)
            # for each tuple of (input/output, commodity) flows:
            #  - build a flow variable which is bounded within the min and max flow
            #  - enforce a piecewise_linear constraint between the node state and the flow
            #  - constraint the respective flow "node.in_flows/out_flows[commodity]" to be node.switch * flow
            #    (i.e., if the machine is on then the final flow is equal to "flow", otherwise it is zero)
            for key, var in [('input', node.in_flows), ('output', node.out_flows)]:
                for commodity, values in self._setpoint[key].items():
                    # create the flow variable and add it to the node
                    v_min, v_max = values.min(), values.max()
                    flow = pyo.Var(domain=pyo.NonNegativeReals, bounds=(v_min, v_max))
                    node.add_component(f'{key}_{commodity}_flow', flow)
                    # create the piecewise linear constraint so that it:
                    #  - stores the output in the "flow" variable
                    #  - takes as input the "node.state" variable
                    #  - is defined via the setpoint index (breakpoints) and the respective flows (values)
                    #  - require tight equality bounds and is modelled using the SOS2 formulation
                    pwl_cst = Piecewise(
                        flow,
                        node.state,
                        pw_pts=list(breakpoints),
                        f_rule=list(values),
                        pw_constr_type='EQ',
                        pw_repn='SOS2'
                    )
                    node.add_component(f'{key}_{commodity}_pwl_cst', pwl_cst)
                    # constraint the final flow (var[commodity]) by linearizing the product node.switch * flow, i.e.:
                    #  - var[commodity] is upper-bounded by the actual value of the flow
                    #  - var[commodity] is greater than flow if node.switch = 1, otherwise the constraint is trivial
                    #  - var[commodity] is lower than zero if node.switch = 0, otherwise the constraint is trivial
                    flow_cst_ub = pyo.Constraint(rule=var[commodity] <= flow)
                    flow_cst_on = pyo.Constraint(rule=var[commodity] >= flow - (1 - node.switch) * v_max)
                    flow_cst_off = pyo.Constraint(rule=var[commodity] <= node.switch * v_max)
                    node.add_component(f'{key}_{commodity}_flow_cst_ub', flow_cst_ub)
                    node.add_component(f'{key}_{commodity}_flow_cst_on', flow_cst_on)
                    node.add_component(f'{key}_{commodity}_flow_cst_off', flow_cst_off)
        return node

    def update(self, rng: np.random.Generator, flows: Dict[Any, Flow], states: Dict[Any, State]):
        self._info['current_state'] = states[self]

    def step(self, flows: Dict[Any, Flow], states: Dict[Any, State]):
        state = states[self]
        # compute total input and output flows from respective edges
        machine_flows = {('input', commodity): 0.0 for commodity in self.commodities_in}
        machine_flows.update({('output', commodity): 0.0 for commodity in self.commodities_out})
        for edge, flow in flows.items():
            if edge.source == self.name:
                machine_flows[('output', edge.commodity)] += flow
            if edge.destination == self.name:
                machine_flows[('input', edge.commodity)] += flow
        # if the state is nan, check that the input/output flows are null
        if np.isnan(state):
            for (key, commodity), flow in machine_flows.items():
                assert np.isclose(flow, 0.0, atol=self.eps), \
                    f"Got non-zero {key} flow {flow} for '{commodity}' despite null setpoint for machine '{self.name}'"
            self._states.append(np.nan)
            return
        # if discrete setpoint
        #  - check that the given state is valid
        #  - check that the flows match the given state
        if self.discrete_setpoint:
            assert state in self._setpoint.index, f"Unsupported state {state} for machine '{self.name}'"
            for (key, commodity), flow in machine_flows.items():
                expected = self._setpoint[(key, commodity)].loc[state]
                assert np.isclose(expected, flow, rtol=self.eps), \
                    f"Flow {expected} expected for machine '{self.name}' with state {state}, got {flow}"
        # if continuous setpoint:
        #  - check that the given state is within the expected bounds
        else:
            lb, ub = self._setpoint.index[[0, -1]]
            assert lb - self.eps <= state <= ub + self.eps, f"Unsupported state {state} for machine '{self.name}'"
            state = np.clip(state, a_min=lb, a_max=ub)
            for (key, commodity), flow in machine_flows.items():
                expected = np.interp(state, xp=self._setpoint.index, fp=self._setpoint[(key, commodity)])
                assert np.isclose(expected, flow, rtol=self.eps), \
                    f"Expected flow {expected} for {key} commodity '{commodity}' in machine '{self.name}', got {flow}"
        # check maximal number of starting by checking that at least one of the following conditions is met:
        #  - the machine is off in this time step
        #  - the machine was on in the previous time step
        #  - the number of starts in the last <t - 1> steps is strictly lower than the maximal value required
        if self.max_starting is not None:
            n, t = self.max_starting
            assert np.isnan(state) or not np.isnan(self.previous_state) or self.starts(t=t - 1) < n, \
                f"Machine '{self.name}' cannot be started for more than {n} times in {t} steps"
        self._states.append(state)
        self._info['current_state'] = None



from abc import ABC
from dataclasses import dataclass
from dataclasses import field
from typing import Callable, Dict, Any
from typing import Set, List, Optional

import numpy as np
import pandas as pd
import pyomo.environ as pyo
# noinspection PyPackageRequirements
from descriptors import classproperty

from powerplantsim.datatypes.node import Node
from powerplantsim.utils.typing import Flow, State


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class ExtremityNode(Node, ABC):
    """A node at the plant extremities that contains a series of values and a variance model for a single commodity."""

    commodity: str = field(kw_only=True)
    """The identifier of the commodity handled by the node."""

    _predictions: np.ndarray = field(kw_only=True)
    """The series of predictions."""

    _variance_fn: Callable[[np.random.Generator, pd.Series], float] = field(kw_only=True)
    """A function f(rng, series) -> variance describing the variance model of true values."""

    _values: List[float] = field(init=False, default_factory=list)
    """The series of actual values, which is filled during the simulation."""

    def __post_init__(self):
        self._info['current_value'] = None
        assert len(self._predictions) == len(self._horizon), \
            f"Predictions should match length of horizon, got {len(self._predictions)} instead of {len(self._horizon)}"

    @classproperty
    def _properties(self) -> List[str]:
        properties = super(ExtremityNode, self)._properties
        return properties + ['commodity']

    @property
    def values(self) -> pd.Series:
        """The series of actual values, which is filled during the simulation."""
        return pd.Series(self._values.copy(), dtype=float, index=self._horizon[:len(self._values)])

    @property
    def current_value(self) -> float:
        """The current value of the node for this time step as computed using the variance model."""
        return self._info['current_value']

    def update(self, rng: np.random.Generator, flows: Dict[Any, Flow], states: Dict[Any, State]):
        # compute the new value as the sum of the prediction and the variance obtained from the variance model
        self._info['current_value'] = self._predictions[self._step] + self._variance_fn(rng, self.values)

    def step(self, flows: Dict[Any, Flow], states: Dict[Any, State]):
        self._values.append(self._info['current_value'])
        self._info['current_value'] = None


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class Client(ExtremityNode, ABC):
    """A node in the plant that buys/asks for a unique commodity."""

    @property
    def commodities_in(self) -> Set[str]:
        return {self.commodity}

    @property
    def commodities_out(self) -> Set[str]:
        return set()


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class Priced(ExtremityNode, ABC):
    """A node in the plant that buys/sells a unique commodity."""

    @classproperty
    def _properties(self) -> List[str]:
        properties = super(Priced, self)._properties
        return properties + ['current_price']

    @property
    def prices(self) -> pd.Series:
        """The series of actual buying prices, which is filled during the simulation."""
        return self.values

    @property
    def current_price(self) -> Optional[float]:
        """The current buying price of the node for this time step as computed using the variance model."""
        return self.current_value

    def to_pyomo(self, mutable: bool = False) -> pyo.Block:
        # start from the default node block
        node = super(Priced, self).to_pyomo(mutable=mutable)
        # add a parameter representing the current price (and initialize it if needed)
        kwargs = dict(mutable=True) if mutable else dict(initialize=self.current_price)
        node.current_price = pyo.Param(domain=pyo.NonNegativeReals, **kwargs)
        return node


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class Customer(Client):
    """A node in the plant that asks for a unique commodity."""

    @classproperty
    def kind(self) -> str:
        return 'customer'

    @classproperty
    def _properties(self) -> List[str]:
        properties = super(Customer, self)._properties
        return properties + ['current_demand']

    @property
    def demands(self) -> pd.Series:
        """The series of actual demands, which is filled during the simulation."""
        return self.values

    @property
    def current_demand(self) -> Optional[float]:
        """The current demand of the node for this time step as computed using the variance model."""
        return self.current_value

    def to_pyomo(self, mutable: bool = False) -> pyo.Block:
        # start from the default node block
        node = super(Customer, self).to_pyomo(mutable=mutable)
        # add a parameter representing the current demand (and initialize it if needed)
        kwargs = dict(mutable=True) if mutable else dict(initialize=self.current_demand)
        node.current_demand = pyo.Param(domain=pyo.NonNegativeReals, **kwargs)
        # constraint the demand so that it matches the input flow of the (unique) commodity
        node.satisfy_demand = pyo.Constraint(rule=node.current_demand == node.in_flows[self.commodity])
        return node

    def step(self, flows: Dict[Any, Flow], states: Dict[Any, State]):
        # check that the flow does not exceed the demand
        demand = self._info['current_value']
        flow = np.sum([flow for edge, flow in flows.items() if edge.destination == self.name])
        assert flow <= demand + self.eps, f"Customer node '{self.name}' can accept at most {demand} units, got {flow}"
        super(Customer, self).step(flows=flows, states=states)


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class Purchaser(Client, Priced):
    """A node in the plant that buys a unique commodity."""

    def __post_init__(self):
        super(Purchaser, self).__post_init__()

    @classproperty
    def kind(self) -> str:
        return 'purchaser'

    def to_pyomo(self, mutable: bool = False) -> pyo.Block:
        node = super(Purchaser, self).to_pyomo(mutable=mutable)
        # compute the cost from the input flow for the (unique) commodity and the price
        node.cost = -node.current_price * node.in_flows[self.commodity]
        return node


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class Supplier(Priced):
    """A node in the plant that can supply a unique commodity."""

    @classproperty
    def kind(self) -> str:
        return 'supplier'

    @property
    def commodities_in(self) -> Set[str]:
        return set()

    @property
    def commodities_out(self) -> Set[str]:
        return {self.commodity}

    def to_pyomo(self, mutable: bool = False) -> pyo.Block:
        node = super(Supplier, self).to_pyomo(mutable=mutable)
        # compute the cost from the output flow for the (unique) commodity and the price
        node.cost = node.current_price * node.out_flows[self.commodity]
        return node



import copy
from typing import Optional, Dict, Tuple, Callable, Iterable, Set, List, Any, Union, Literal

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import pyomo.environ as pyo
from tqdm import tqdm

from powerplantsim import utils
from powerplantsim.datatypes import Node, Machine, Supplier, SingleEdge, Storage, Purchaser, Customer, ExtremityNode, \
    Edge
from powerplantsim.plant import drawing, execution
from powerplantsim.plant.action import DefaultRecourseAction, CallableRecourseAction, RecourseAction
from powerplantsim.plant.callback import Callback
from powerplantsim.plant.execution import check_plan
from powerplantsim.utils.typing import Plan, EdgeID


class Plant:
    """Defines a power plant based on its topology, involved commodities, and predicted prices and demands."""

    def __init__(self, horizon: Union[int, Iterable[float]], seed: int = 0, name: Optional[str] = None):
        """
        :param horizon:
            The time horizon of the simulation.
            If an integer is passed, the index will be {0, ..., horizon - 1}.
            Otherwise, an explicit list, numpy array or pandas series can be passed.

        :param seed:
            The seed used for random number operations.

        :param name:
            The name of the plant.
        """
        # convert horizon to a standard format (pd.Index)
        if isinstance(horizon, int):
            assert horizon > 0, f"The time horizon must be a strictly positive integer, got {horizon}"
            horizon = np.arange(horizon)
        horizon = pd.Index(horizon)

        self._name: str = hex(id(self)).upper() if name is None else name
        self._rng: np.random.Generator = np.random.default_rng(seed=seed)
        self._horizon: pd.Index = horizon
        self._commodities: Set[str] = set()
        self._nodes: Dict[str, Set[Node]] = dict()
        self._edges: Set[SingleEdge] = set()
        self._step: int = -1

    @property
    def name(self) -> str:
        """The name of the plant."""
        return self._name

    @property
    def step(self) -> Optional[int]:
        """The current step of the simulation, or None if the simulation is not started yet."""
        return None if self._step == -1 else self._step

    @property
    def index(self) -> Optional:
        """The current index of the simulation as in the time horizon, or None if the simulation is not started."""
        return None if self._step == -1 else self._horizon[self._step]

    @property
    def horizon(self) -> pd.Index:
        """A pandas series representing the time index of the simulation."""
        return self._horizon.copy()

    @property
    def commodities(self) -> Set[str]:
        return {c for c in self._commodities}

    @property
    def suppliers(self) -> Dict[str, Supplier]:
        """The supplier nodes in the plant."""
        # noinspection PyTypeChecker
        return {s.name: s for s in self._nodes.get(Supplier.kind, set())}

    @property
    def customers(self) -> Dict[str, Customer]:
        """The customer nodes in the plant."""
        # noinspection PyTypeChecker
        return {c.name: c for c in self._nodes.get(Customer.kind, set())}

    @property
    def purchasers(self) -> Dict[str, Purchaser]:
        """The purchaser nodes in the plant."""
        # noinspection PyTypeChecker
        return {c.name: c for c in self._nodes.get(Purchaser.kind, set())}

    @property
    def machines(self) -> Dict[str, Machine]:
        """The machine nodes in the plant."""
        # noinspection PyTypeChecker
        return {m.name: m for m in self._nodes.get(Machine.kind, set())}

    @property
    def storages(self) -> Dict[str, Storage]:
        """The storage nodes in the plant."""
        # noinspection PyTypeChecker
        return {s.name: s for s in self._nodes.get(Storage.kind, set())}

    def nodes(self, indexed: bool = False) -> Dict[str, Union[Node, Dict[str, Node]]]:
        """Returns the nodes in the plant.

        :param indexed:
            Whether to index the dictionary by type.

        :return:
            Either a dictionary {type: {name: node}} or a simple dictionary {name: node}.
        """
        if indexed:
            return {kind: {n.name: n for n in nodes} for kind, nodes in self._nodes.items()}
        else:
            return {n.name: n for nodes in self._nodes.values() for n in nodes}

    def edges(self,
              sources: Union[None, str, Iterable[str]] = None,
              destinations: Union[None, str, Iterable[str]] = None,
              commodities: Union[None, str, Iterable[str]] = None) -> Dict[EdgeID, Edge]:
        """Returns the edges in the plant indexed either via nodes, or via nodes and commodity.

        :param sources:
            The identifier of the source nodes, used to filter just the edges starting from this node.

        :param destinations:
            The identifier of the destination nodes, used to filter just the edges ending in this node.

        :param commodities:
            The identifier of the commodities, used to filter just the edges ending in this node.

        :return:
            A dictionary <(source, destination), edge> with nodes pairs as key and edge info as value.
        """
        # get filtering functions for destinations and commodities
        check_sour = utils.get_filtering_function(user_input=sources)
        check_dest = utils.get_filtering_function(user_input=destinations)
        check_edge = utils.get_filtering_function(user_input=commodities)
        # build data structure containing all the necessary information
        return {
            e.key: e
            for e in self._edges
            if check_sour(e.source) and check_dest(e.destination) and check_edge(e.commodity)
        }

    def __repr__(self) -> str:
        output = f"Plant('{self._name}'):\n"
        output += f"  > Horiz: {[step for step in self._horizon.values]}\n"
        output += f"  > Nodes: {[node for node in self.nodes().values()]}\n"
        output += f"  > Edges: {[edge for edge in self.edges().keys()]}"
        return output

    def graph(self, attributes: bool = False) -> nx.DiGraph:
        """Builds the graph representing the power plant.

        :param attributes:
            Whether to include node/edge attributes.

        :return:
            A networkx DiGraph object representing the power plant.
        """
        g = nx.DiGraph()
        if attributes:
            for name, node in self.nodes().items():
                g.add_node(name, **node.dict)
            for edge in self._edges:
                g.add_edge(edge.source, edge.destination, **edge.dict)
        else:
            g.add_nodes_from(self.nodes().keys())
            g.add_edges_from(self.edges().keys())
        return g

    def to_pyomo(self, mutable: bool = False) -> pyo.ConcreteModel:
        """Encodes the plant as a Pyomo model.

        :param mutable:
            Whether to set simulation-specific parameters (e.g., current_price/current_demand/etc.) as mutable and
            without an initialization value, or to initialize them based on their specific value of the current step.

        :return:
            A pyomo.environment.ConcreteModel object modeling the plant.
        """
        # build model and define data sets for nodes and commodities
        model = pyo.ConcreteModel(name='plant')
        model.nodes = pyo.Set(initialize=[node.name for nodes in self._nodes.values() for node in nodes])
        model.commodities = pyo.Set(initialize=list(self._commodities))

        # create components for datatypes and add them to the model
        edges = {(e.source, e.destination, e.commodity): e.to_pyomo(mutable=mutable) for e in self._edges}
        nodes = {n.name: n.to_pyomo(mutable=mutable) for nodes in self._nodes.values() for n in nodes}
        for component in [*nodes.values(), *edges.values()]:
            model.add_component(component.name, component)

        # add constraints between edges variables and nodes variables
        #  - compute the input and output flows for pairs (node, commodity) as a sum of edges flows
        #  - then, use the cross product "nodes x commodities" to both the computed flow sums and the nodes' flows
        #  - finally, use these two variables (flow sums and nodes' flows) to impose an equality constraint
        in_flows = {(destination, commodity): 0.0 for _, destination, commodity in edges.keys()}
        out_flows = {(source, commodity): 0.0 for source, _, commodity in edges.keys()}
        for (source, destination, commodity), edge in edges.items():
            # noinspection PyUnresolvedReferences
            in_flows[(destination, commodity)] += edge.flow
            # noinspection PyUnresolvedReferences
            out_flows[(source, commodity)] += edge.flow

        # noinspection PyUnresolvedReferences
        @model.Constraint(model.nodes * model.commodities)
        def in_constraints(_, destination, commodity):
            flows = in_flows.get((destination, commodity))
            return pyo.Constraint.Skip if flows is None else flows == nodes[destination].in_flows[commodity]

        # noinspection PyUnresolvedReferences
        @model.Constraint(model.nodes * model.commodities)
        def out_constraints(_, source, commodity):
            flows = out_flows.get((source, commodity))
            return pyo.Constraint.Skip if flows is None else flows == nodes[source].out_flows[commodity]

        # eventually return the model
        return model

    def copy(self):
        """Copies the plant object.

        :return:
            A copy of the plant object.
        """
        return copy.deepcopy(self)

    def _check_and_update(self,
                          node: Node,
                          parents: Union[None, str, Iterable[str]],
                          commodity: Optional[str],
                          min_flow: Optional[float],
                          max_flow: Optional[float]):
        # check that the node has a unique identifier and append it to the designed internal data structure
        for kind, nodes in self._nodes.items():
            assert node not in nodes, f"There is already a {kind} node '{node.name}', please use another identifier"
        # if the node is not a source (supplier), retrieve the node parent and check that is exists
        edges = []
        if parents is not None:
            parents = [parents] if isinstance(parents, str) else parents
            assert len(parents) > 0, f"{node.kind.title()} node must have at least one parent"
            for name in parents:
                parent = self.nodes().get(name)
                assert parent is not None, f"Parent node '{name}' has not been added yet"
                # create an edge instance using the parent as source and the new node as destination
                edge = SingleEdge(
                    _plant=self,
                    _source=parent,
                    _destination=node,
                    commodity=commodity,
                    min_flow=min_flow,
                    max_flow=max_flow
                )
                edges.append(edge)
        # once all the checks have been passed, add the information to the plant
        # add the new commodities to the set
        self._commodities.update(node.commodities_in)
        self._commodities.update(node.commodities_out)
        # add the node to its respective set
        node_set = self._nodes.get(node.kind, set())
        node_set.add(node)
        self._nodes[node.kind] = node_set
        # add the edges
        for edge in edges:
            self._edges.add(edge)

    def add_extremity(self,
                      kind: Literal['customer', 'purchaser', 'supplier'],
                      name: str,
                      commodity: str,
                      predictions: Union[float, Iterable[float]],
                      variance: Callable[[np.random.Generator, pd.Series], float] = lambda rng, series: 0.0,
                      parents: Union[None, str, Iterable[str]] = None) -> ExtremityNode:
        """Adds an extremity node (supplier, client, purchaser) to the plant topology.

        :param kind:
            The kind of extremity node, either 'customer', 'purchaser', or 'supplier'.

        :param name:
            The name of the extremity node.

        :param commodity:
            The identifier of the commodity that it supplies, which must have been registered before.

        :param predictions:
            The vector of predicted selling prices for the commodity during the time horizon, or a float if the
            predictions are constant throughout the simulation. If an iterable is passed, it must have the same length
            of the time horizon.

        :param variance:
            A function f(<rng>, <series>) -> <variance> which defines the variance model of the true prices
            respectively to the predictions.
            The random number generator is used to get reproducible results, while the input series represents the
            vector of previous values indexed by the datetime information passed to the plant; the last value of the
            series is always a nan value, and it will be computed at the successive iteration based on the respective
            predicted price and the output of this function.
            Indeed, the function must return a real number <eps> which represents the delta between the predicted and
            the true price; for an input series with length L, the true price will be eventually computed as:
                true = self.prices[L] + <eps>

        :param parents:
            The identifier of the parent nodes that are connected with the input of this extremity node, or None in
            case kind == 'supplier'.

        :return:
            The added extremity node.
        """
        if isinstance(predictions, float):
            predictions = np.ones_like(self._horizon) * predictions
        else:
            predictions = np.array(predictions)
        # create an internal supplier node and add it to the internal data structure and the graph
        if kind == 'supplier':
            assert parents is None, f"Supplier node {name} cannot accept parents"
            node = Supplier(
                _plant=self,
                name=name,
                commodity=commodity,
                _predictions=predictions,
                _variance_fn=variance
            )
        elif kind == 'purchaser':
            assert parents is not None, f"Purchaser node {name} must have parents"
            node = Purchaser(
                _plant=self,
                name=name,
                commodity=commodity,
                _predictions=predictions,
                _variance_fn=variance
            )
        elif kind == 'customer':
            assert parents is not None, f"Customer node {name} must have parents"
            node = Customer(
                _plant=self,
                name=name,
                commodity=commodity,
                _predictions=predictions,
                _variance_fn=variance
            )
        else:
            raise AssertionError(f"Unknown extremity node kind {kind}")
        self._check_and_update(
            node=node,
            parents=parents,
            commodity=commodity,
            min_flow=0.0,
            max_flow=float('inf')
        )
        return node

    def add_machine(self,
                    name: str,
                    parents: Union[str, Iterable[str]],
                    commodity: str,
                    setpoint: Iterable[float],
                    inputs: Iterable[float],
                    outputs: Dict[str, Iterable[float]],
                    discrete: bool = False,
                    max_starting: Optional[Tuple[int, int]] = None,
                    cost: float = 0.0,
                    min_flow: float = 0.0,
                    max_flow: float = float('inf')) -> Machine:
        """Adds a machine node to the topology.

        :param name:
            The name of the machine node.

        :param parents:
            The identifier of the parent nodes that are connected with the input of this machine node.

        :param commodity:
            The input commodity of the machine.

        :param setpoint:
            The setpoint of the machine.

        :param inputs:
            The amount of input commodity flow paired with the setpoint.

        :param outputs:
            A dictionary {<output_commodity_i>: <output_flow_i>} where <output_commodity_i> is the name of each output
            commodity generated by the machine and <output_flow_i> is the list of respective flows.

        :param discrete:
            Whether the setpoint is discrete or continuous.

        :param max_starting:
            A tuple <N, T> where N is the maximal number of times that the machine can be switched on in T units.

        :param cost:
            The cost to operate the machine.

        :param min_flow:
            The minimal flow of commodity that can pass in the edge.

        :param max_flow:
            The maximal flow of commodity that can pass in the edge.

        :return:
            The added machine node.
        """
        # convert setpoint to a standard format (pd.Series)
        index = pd.Index(setpoint, dtype=float, name='setpoint')
        input_flows = pd.DataFrame({commodity: inputs}, dtype=float, index=index)
        output_flows = pd.DataFrame({c: f for c, f in outputs.items()}, dtype=float, index=index)
        setpoint = pd.concat((input_flows, output_flows), keys=['input', 'output'], axis=1)
        # create an internal machine node and add it to the internal data structure and the graph
        machine = Machine(
            _plant=self,
            name=name,
            _setpoint=setpoint,
            discrete_setpoint=discrete,
            max_starting=max_starting,
            cost=cost
        )
        self._check_and_update(
            node=machine,
            parents=parents,
            commodity=commodity,
            min_flow=min_flow,
            max_flow=max_flow
        )
        return machine

    def add_storage(self,
                    name: str,
                    commodity: str,
                    capacity: float,
                    parents: Union[str, List[str]],
                    dissipation: float = 0.0,
                    rates: Union[None, float, Tuple[float, float]] = None,
                    min_flow: float = 0.0,
                    max_flow: float = float('inf')) -> Storage:
        """Adds a storage node to the topology.

        :param name:
            The name of the storage node.

        :param commodity:
            The commodity stored by the storage node.

        :param capacity:
            The storage capacity, which must be a strictly positive and finite number.

        :param parents:
            The identifier of the parent nodes that are connected with the input of this storage node.

        :param dissipation:
            The dissipation rate of the storage at every time unit, which must be a float in [0, 1].

        :param rates:
            Either a tuple (charge_rate, discharge_rate) containing the maximal charge rate (input flow) and discharge
            rate (output flow) in a time unit, or a single float value in case charge_rate == discharge_rate. If None
            is passed, the charge and discharge rates are set to capacity.

        :param min_flow:
            The minimal flow of commodity that can pass in the edge.

        :param max_flow:
            The maximal flow of commodity that can pass in the edge.

        :return:
            The added storage node.
        """
        # handle rates
        if rates is None:
            charge_rate, discharge_rate = capacity, capacity
        elif isinstance(rates, float):
            charge_rate, discharge_rate = rates, rates
        else:
            charge_rate, discharge_rate = rates
        # create an internal machine node and add it to the internal data structure and the graph
        storage = Storage(
            _plant=self,
            name=name,
            commodity=commodity,
            capacity=capacity,
            dissipation=dissipation,
            charge_rate=charge_rate,
            discharge_rate=discharge_rate
        )
        self._check_and_update(
            node=storage,
            parents=parents,
            commodity=commodity,
            min_flow=min_flow,
            max_flow=max_flow
        )
        return storage

    def draw(self,
             figsize: Tuple[int, int] = (16, 9),
             node_pos: Union[str, List[Iterable[Optional[str]]], Dict[str, Any]] = 'lp',
             node_colors: Union[None, str, Dict[str, str]] = None,
             node_markers: Union[None, str, Dict[str, str]] = None,
             edge_colors: Union[str, Dict[str, str]] = 'black',
             edge_styles: Union[str, Dict[str, str]] = 'solid',
             node_size: float = 30,
             edge_width: float = 2,
             legend: Optional[int] = 20):
        """Draws the plant topology.

        :param figsize:
            The matplotlib figsize parameter.

        :param node_pos:
            If the string 'sp' is passed, arranges the nodes into layers using breadth first search to get the shortest
            path. If the string 'lp' is passed, arranges the nodes into layers using negative unitary cost to get the
            longest  path. If a list is passed, it is considered as a mapping <layer: nodes> representing in which
            layer to display the nodes (None values can be used to add placeholder nodes). If a dictionary is passed,
            it is considered as a mapping <node: position> representing where exactly to display each node.

        :param node_colors:
            Either a string representing the color of the nodes, or a dictionary {kind: color} which associates a color
            to each node kind ('supplier', 'client', 'machine').

        :param node_markers:
            Either a string representing the shape of the nodes, or a dictionary {kind: shape} which associates a shape
            to each node kind ('supplier', 'client', 'machine', 'storage').

        :param edge_colors:
            Either a string representing the color of the edges, or a dictionary {commodity: color} which associates a
            color to each commodity that flows in an edge.

        :param edge_styles:
            Either a string representing the style of the edges, or a dictionary {commodity: style} which associates a
            style to each commodity that flows in an edge.

        :param node_size:
            The size of the nodes.

        :param edge_width:
            The width of the edges and of the node's borders.

        :param legend:
            The size of the legend, or None for no legend.
        """
        # retrieve plant info, build the figure, and compute node positions
        nodes = self.nodes(indexed=True)
        graph = self.graph(attributes=False)
        _, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize, tight_layout=True)
        pos = drawing.get_node_positions(graph=graph, sources=nodes['supplier'].keys(), node_pos=node_pos)
        # retrieve nodes' styling information and draw them accordingly
        labels = []
        styles = drawing.get_node_style(colors=node_colors, markers=node_markers)
        for kind, node_list in nodes.items():
            drawing.draw_nodes(
                graph=graph,
                pos=pos,
                nodes=node_list.keys(),
                style=styles[kind],
                size=node_size,
                width=edge_width,
                ax=ax
            )
            handler = drawing.build_node_label(kind=kind, style=styles[kind])
            labels.append(handler)
        # retrieve edges' styling information and draw them accordingly
        styles = drawing.get_edge_style(colors=edge_colors, shapes=edge_styles, commodities=list(self._commodities))
        edges = pd.DataFrame([{'commodity': e.commodity, 'edge': e} for e in self.edges().values()])
        for commodity, edge_list in edges.groupby('commodity'):
            commodity = str(commodity)
            drawing.draw_edges(
                graph=graph,
                pos=pos,
                edges={e.key for e in edge_list['edge']},
                style=styles[commodity],
                size=node_size,
                width=edge_width,
                ax=ax
            )
            handler = drawing.build_edge_label(commodity=commodity, style=styles[commodity])
            labels.append(handler)
        # plot the legend if necessary, and eventually show the result
        if legend is not None:
            ax = plt.legend(handles=labels, prop={'size': legend})
            for handle in ax.legend_handles:
                # noinspection PyUnresolvedReferences
                handle.set_markersize(legend)
        plt.show()

    def run(self,
            plan: Union[Plan, pd.DataFrame],
            action: Union[str, RecourseAction, Callable[[Any], Plan]],
            callbacks: Optional[Iterable[Callback]] = None,
            progress: bool = True) -> execution.SimulationOutput:
        """Runs a simulation up to the time horizon using the given plan.

        :param plan:
            The energetic plan of the power plant defined as vectors of edge flows and machine states. It can be a
            dictionary <machine | edge: states | flows> that maps each machine (indexed by its name) to either a fixed
            state or an iterable of such which should match the time horizon of the simulation, and each edge (indexed
            by the tuple of source and destination) to either a fixed flow or an iterable of such which should match
            the time horizon of the simulation. Otherwise, it can be a dataframe where each column is indexed by the
            machine name of the edge key, and the value is a vector of states or flows, respectively, with the index
            of the dataframe that matches the time horizon of the simulation.

        :param action:
            If a RecourseAction object is passed, it is used to build the real plan at each step.
            If a function f(plant) -> plan, this is wrapped into a CallableRecourseAction object.
            If a string is passed, this is interpreted as the solver to use in the default greedy recourse action.

        :param callbacks:
            A list of Callbacks instances that are called throughout the simulation.

        :param progress:
            Whether to show a tqdm progressbar throughout the simulation.

        :return:
            A SimulationOutput object containing all the information about true prices, demands, setpoints, and storage.
        """
        assert self._step == -1, "Simulation for this plant was already run, create a new instance to run another one"
        execution.check_plant(plant=self)
        callbacks = [] if callbacks is None else callbacks
        # handle recourse action (if None use default action, if Callable build a custom recourse action)
        if isinstance(action, str):
            action = DefaultRecourseAction(solver=action)
        elif isinstance(action, Callable):
            action = CallableRecourseAction(action=action)
        action.build(plant=self)
        # retrieve data and process input plan
        nodes = self.nodes()
        edges = self.edges()
        machines = self.machines
        datatypes = {**edges, **nodes}
        plan, states, flows = execution.process_plan(plan=plan, machines=machines, edges=edges, horizon=self._horizon)
        # run callbacks before simulation start
        for callback in callbacks:
            callback.on_simulation_start(plant=self, states=states, flows=flows)
        # start the simulation
        for row in tqdm(plan, desc='Simulation Status') if progress else plan:
            self._step += 1
            # update the simulation objects before the recourse action
            for datatype in datatypes.values():
                datatype.update(rng=self._rng, states=row.states, flows=row.flows)
            # run callbacks on iteration start
            for callback in callbacks:
                callback.on_iteration_update(plant=self)
            # compute the updated flows and states using the recourse action
            updated_states, updated_flows = check_plan(
                plan=action.execute(),
                machines=machines,
                edges=edges
            )
            # run callbacks on iteration update
            for callback in callbacks:
                callback.on_iteration_recourse(plant=self, states=updated_states, flows=updated_flows)
            # update the simulation objects after the recourse action
            for datatype in datatypes.values():
                datatype.step(flows=updated_flows, states=updated_states)
            # run callbacks on iteration end
            for callback in callbacks:
                callback.on_iteration_step(plant=self)
        output = execution.build_output(nodes=nodes.values(), edges=edges.values(), horizon=self._horizon)
        # run callbacks on simulation end
        for callback in callbacks:
            callback.on_simulation_end(plant=self)
        return output



from dataclasses import dataclass
from typing import Any, Dict, Tuple, List, Iterable, Union

NodeID = str
"""Datatype for node identifier, i.e, its name."""

SingleEdgeID = Tuple[str, str]
"""Datatype for single-edge identifier, i.e., a tuple (source, destination)."""

MultiEdgeID = Tuple[str, str, str]
"""Datatype for multi-edge identifier, i.e., a tuple (source, destination, commodity)."""

EdgeID = SingleEdgeID
"""Alias for plant edges identifier, since plant do not support multi-graphs yet."""

State = float
"""Datatype for a single machine state specification (np.nan for machine off)."""

Flow = float
"""Datatype for a single edge flow specification."""

StepPlan = Dict[Union[NodeID, EdgeID], Union[State, Flow]]
"""Datatype for the specification of the plan of a single step, i.e., a dictionary <machine | edge: state | flow>."""

Plan = Dict[Union[NodeID, EdgeID], Union[State, Flow, Iterable[State], Iterable[Flow]]]
"""Datatype for specification of the total plan, i.e., a dictionary <machine | edge: state/states | flow/flows>."""


# noinspection PyUnresolvedReferences
@dataclass(frozen=True, unsafe_hash=True, slots=True)
class NamedTuple:
    """Template dataclass that models a named tuple."""

    def __getitem__(self, item: Any):
        if isinstance(item, int):
            item = self.__slots__[item]
        return getattr(self, item)

    def __len__(self):
        return len(self.__slots__)

    def __eq__(self, other: Any):
        # if another NamedTuple is passed, create
        if isinstance(other, NamedTuple):
            return self.dict == other.dict
        else:
            return self.tuple == other

    @property
    def dict(self) -> Dict[str, Any]:
        return {param: getattr(self, param) for param in self.__slots__}

    @property
    def list(self) -> List[Any]:
        return [getattr(self, param) for param in self.__slots__]

    @property
    def tuple(self) -> Tuple:
        return tuple(self.list)



from powerplantsim.datatypes.edge import Edge, SingleEdge, MultiEdge
from powerplantsim.datatypes.extremity import ExtremityNode, Client, Priced, Customer, Purchaser, Supplier
from powerplantsim.datatypes.machine import Machine
from powerplantsim.datatypes.node import Node
from powerplantsim.datatypes.storage import Storage



from powerplantsim.utils.matching import get_filtering_function, get_matching_object, get_indexed_object
from powerplantsim.utils.strings import stringify
from powerplantsim.utils.typing import NamedTuple



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Set, List, Tuple

import pyomo.environ as pyo
# noinspection PyPackageRequirements
from descriptors import classproperty

from powerplantsim.datatypes.datatype import DataType
from powerplantsim.utils.typing import NodeID


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True, slots=True)
class Node(DataType, ABC):
    """Data class for an abstract node in the plant."""

    name: str = field(kw_only=True)
    """The name of the datatype."""

    @classproperty
    @abstractmethod
    def kind(self) -> str:
        """The node type."""
        pass

    @classproperty
    def _properties(self) -> List[str]:
        properties = super(Node, self)._properties
        return properties + ['kind']

    @property
    @abstractmethod
    def commodities_in(self) -> Set[str]:
        """The set of input commodities that is accepted."""
        pass

    @property
    @abstractmethod
    def commodities_out(self) -> Set[str]:
        """The set of output commodities that is returned."""
        pass

    @property
    def key(self) -> NodeID:
        return self.name

    @property
    def _edges(self) -> Tuple[set, set]:
        """A tuple <in_edges, out_edges> of sets of Edge objects containing all the input/output edges of this node."""
        in_edges, out_edges = set(), set()
        for edge in self._plant.edges().values():
            if edge.source == self.name:
                out_edges.add(edge)
            elif edge.destination == self.name:
                in_edges.add(edge)
        return in_edges, out_edges

    def _instance(self, other) -> bool:
        return isinstance(other, Node)

    def to_pyomo(self, mutable: bool = False) -> pyo.Block:
        # build a node block with two variable arrays representing the input/output flows indexed by commodity
        node = pyo.Block(concrete=True, name=self.name)
        # these flows are not bounded, but they must be constrained to be equal to the sum of edge variables
        node.in_flows = pyo.Var(self.commodities_in, domain=pyo.NonNegativeReals)
        node.out_flows = pyo.Var(self.commodities_out, domain=pyo.NonNegativeReals)
        return node



import inspect
from types import LambdaType, FunctionType, MethodType


def stringify(value, prefix: str = '', suffix: str = '') -> str:
    """Custom method to handle convert certain data types to strings.

    :param value:
        Any object.

    :param prefix:
        A prefix in the final string version; default = ''.

    :param suffix:
        A suffix in the final string version; default = ''.

    :return:
        The string version of the given object.
    """
    if value is None:
        return ''
    elif isinstance(value, dict):
        elements = [stringify(k) + ': ' + stringify(v) for k, v in value.items()]
        string = f"{{{', '.join(elements)}}}"
    elif isinstance(value, (set, list, tuple)):
        delimiter = '({})' if isinstance(value, tuple) else ('[{}]' if isinstance(value, list) else '{{{}}}')
        string = delimiter.format(', '.join([stringify(v) for v in value]))
    elif isinstance(value, (LambdaType, FunctionType, MethodType)):
        # include the parameters in the function representation
        string = f"{value.__name__}{inspect.signature(value)}"
    else:
        string = value.__name__ if hasattr(value, '__name__') else f"{value!r}"
    return f'{prefix}{string}{suffix}'



from abc import abstractmethod
from typing import Dict

import pandas as pd

from powerplantsim.datatypes import Edge, Node
from powerplantsim.utils.typing import State, Flow


class Callback:
    """Basic callback interface for a simulation."""

    @abstractmethod
    def on_simulation_start(self, plant, states: Dict[Node, pd.Series], flows: Dict[Edge, pd.Series]):
        """Callback hook that is called before the first step of the simulation.

        :param plant:
            The plant object on which the callback is attached.

        :param states:
            The input states of the whole plan indexed by node.

        :param flows:
            The input flows of the whole plan indexed by edge.
        """
        pass

    @abstractmethod
    def on_iteration_update(self, plant):
        """Callback hook that is called soon after the datatypes are updated with the given plan.

        :param plant:
            The plant object on which the callback is attached.
        """
        pass

    @abstractmethod
    def on_iteration_recourse(self, plant, states: Dict[Node, State], flows: Dict[Edge, Flow]):
        """Callback hook that is called soon after the actual plan is computed by the recourse action.

        :param plant:
            The plant object on which the callback is attached.

        :param states:
            The input states of the given iteration indexed by node.

        :param flows:
            The input flows of the given iteration indexed by edge.
        """
        pass

    @abstractmethod
    def on_iteration_step(self, plant):
        """Callback hook that is called soon after the datatypes are updated with the actual plan.

        :param plant:
            The plant object on which the callback is attached.
        """
        pass

    @abstractmethod
    def on_simulation_end(self, plant):
        """Callback hook that is called once the simulation is ended.

        :param plant:
            The plant object on which the callback is attached.
        """
        pass



from abc import ABC
from dataclasses import dataclass, field
from typing import List, Optional, Tuple, Dict, Any

import numpy as np
import pandas as pd
import pyomo.environ as pyo
# noinspection PyPackageRequirements
from descriptors import classproperty

from powerplantsim.datatypes.datatype import DataType
from powerplantsim.datatypes.node import Node
from powerplantsim.utils.typing import SingleEdgeID, Flow, MultiEdgeID, State


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class Edge(DataType, ABC):
    """An (abstract) edge in the plant."""

    _source: Node = field(kw_only=True)
    """The source node."""

    _destination: Node = field(kw_only=True)
    """The destination node."""

    commodity: str = field(kw_only=True)
    """The type of commodity that flows in the edge, which can be uniquely determined by the destination input."""

    min_flow: float = field(kw_only=True)
    """The minimal flow of commodity."""

    max_flow: float = field(kw_only=True)
    """The maximal flow of commodity."""

    _flows: List[float] = field(init=False, default_factory=list)
    """The series of actual flows, which is filled during the simulation."""

    def __post_init__(self):
        self._info['current_flow'] = None
        assert self.min_flow >= 0, f"The minimum flow cannot be negative, got {self.min_flow}"
        assert self.max_flow >= self.min_flow, \
            f"The maximum flow cannot be lower than the minimum, got {self.max_flow} < {self.min_flow}"
        assert self.commodity in self._source.commodities_out, \
            f"Source node '{self._source.name}' should return commodity '{self.commodity}', " \
            f"but it returns {self._source.commodities_out}"
        assert self.commodity in self._destination.commodities_in, \
            f"Destination node '{self._destination.name}' should accept commodity '{self.commodity}', " \
            f"but it accepts {self._destination.commodities_in}"

    @classproperty
    def _properties(self) -> List[str]:
        properties = super(Edge, self)._properties
        return properties + ['source', 'destination', 'commodity', 'min_flow', 'max_flow', 'bounds', 'current_flow']

    @property
    def source(self) -> str:
        """The source node."""
        return self._source.name

    @property
    def destination(self) -> str:
        """The destination node."""
        return self._destination.name

    @property
    def bounds(self) -> Tuple[float, float]:
        """The lower and upper flow bounds for the commodity."""
        return self.min_flow, self.max_flow

    @property
    def flows(self) -> pd.Series:
        """The series of actual flows, which is filled during the simulation."""
        return pd.Series(self._flows, dtype=float, index=self._horizon[:len(self._flows)])

    @property
    def current_flow(self) -> Optional[Flow]:
        """The current flow on the edge for this time step as provided by the user."""
        return self._info['current_flow']

    def to_pyomo(self, mutable: bool = False) -> pyo.Block:
        # build an edge block with a variable representing the flow
        edge = pyo.Block(concrete=True, name=self.name)
        # the variable is bounded by the min/max bounds and is initialized to the current flow if needed
        kwargs = dict() if mutable else dict(initialize=self.current_flow)
        edge.flow = pyo.Var(domain=pyo.NonNegativeReals, bounds=self.bounds, **kwargs)
        return edge

    def update(self, rng: np.random.Generator, flows: Dict[Any, Flow], states: Dict[Any, State]):
        self._info['current_flow'] = flows[self]

    def step(self, flows: Dict[Any, Flow], states: Dict[Any, State]):
        flow = flows[self]
        assert flow >= self.min_flow - self.eps, f"Flow for edge {self.key} should be >= {self.min_flow}, got {flow}"
        assert flow <= self.max_flow + self.eps, f"Flow for edge {self.key} should be <= {self.max_flow}, got {flow}"
        self._flows.append(np.clip(flow, a_min=self.min_flow, a_max=self.max_flow))
        self._info['current_flow'] = None


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class SingleEdge(Edge):
    """An edge in a plant where two nodes can be connected by a unique edge (i.e., a graph)."""

    @property
    def key(self) -> SingleEdgeID:
        return self.source, self.destination

    @property
    def name(self) -> str:
        return f"{self.source} --> {self.destination}"


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class MultiEdge(Edge):
    """An edge in a plant where two nodes can be connected by a multiple edges (i.e., a multi-graph)."""

    @property
    def key(self) -> MultiEdgeID:
        return self.source, self.destination, self.commodity

    @property
    def name(self) -> str:
        return f"{self.source} --({self.commodity})--> {self.destination}"



from abc import abstractmethod
from typing import Dict, Tuple, Union, Optional, Callable, Any

import numpy as np
import pyomo.environ as pyo

from powerplantsim.utils.typing import Plan


class RecourseAction:
    """Abstract class that defines the recourse action to apply to the predicted plan to generate the actual plan,
    after the true costs and demands are disclosed."""

    def __init__(self):
        self._plant: Optional = None

    def build(self, plant):
        """Builds the recourse action by attaching it to a specific plant.

        :param plant:
            The power plant object.

        :return:
            The RecourseAction instance itself.
        """
        assert self._plant is None, "Recourse action already built."
        self._plant = plant
        return self

    @abstractmethod
    def execute(self) -> Plan:
        """Performs the recourse action on the power plant for a given step.

        :return:
            A dictionary {machine | edge: <updated_state | updated_flow>} where a machine is identified by its name and
            an edge is identified by the tuple of the names of the node that is connecting, while updated_state and
            updated_flow is the value of the actual state/flow.
        """
        pass


class CallableRecourseAction(RecourseAction):
    """A recourse action which is built from a function f(plant) -> plan."""

    def __init__(self, action: Callable[[Any], Plan]):
        """
        :param action:
            A function f(plant) -> plan which performs the recourse action on the power plant for a given step.
        """
        super(CallableRecourseAction, self).__init__()
        self._action: Callable[[Any], Plan] = action

    def execute(self) -> Plan:
        return self._action(self._plant)


class DefaultRecourseAction(RecourseAction):
    """A default, greedy implementation of the recourse action."""

    MachineWeight = Tuple[float, float, float]
    """Type to define the cost of operating a machine, i.e., a tuple (switch_on, switch_off, state)."""

    def __init__(self,
                 solver: str,
                 decimals: int = 9,
                 cost_weight: Optional[float] = 1.0,
                 storage_weight: Union[None, float, Dict[str, float]] = 1.0,
                 machine_weight: Union[None, MachineWeight, Dict[str, MachineWeight]] = (1.0, 1.0, 0.0),
                 **solver_options: Any):
        """
        :param solver:
            The underlying MIP solver to use in Pyomo.

        :param decimals:
            The decimals used for final rounding operations to account for numerical errors in the solving process.

        :param cost_weight:
            The objective coefficient for the total cost of the produced plan.

        :param storage_weight:
            Either a float used as the objective coefficient for the total difference between the predicted and the
            produced plan in terms of storage states, or a dictionary {storage : accumulated_storage_diff} containing
            the objective coefficients of the accumulated storage difference for individual storages.

        :param machine_weight:
            Either a float used as the objective coefficient of the total difference between the predicted and the
            produced plan in terms of machine states, or a dictionary {machine : (switch_on, switch_off, state)}
            containing the objective coefficients of the switch_on, switch_off and state difference for individual
            machines.

        :param solver_options:
            Additional options of the underlying solver.
        """
        super(DefaultRecourseAction, self).__init__()

        self._solver: str = solver
        self._solver_options: Dict[str, Any] = solver_options
        self._decimals: int = decimals
        self._cost_weight: Optional[float] = cost_weight
        self._storage_weight: Union[None, float, Dict[str, float]] = storage_weight
        self._machine_weight: Union[None, float, Dict[str, DefaultRecourseAction.MachineWeight]] = machine_weight

    def build(self, plant):
        super(DefaultRecourseAction, self).build(plant)
        # handle the storage weight input in case a single value is passed instead of dictionary
        if isinstance(self._storage_weight, float):
            self._storage_weight = {s: self._storage_weight for s in self._plant.storages}
        # handle the machine weight input in case a single value is passed instead of dictionary
        if isinstance(self._machine_weight, tuple):
            self._machine_weight = {m: self._machine_weight for m in self._plant.machines}
        return self

    # noinspection PyTypeChecker
    def execute(self) -> Plan:
        # retrieve the pyomo model representing the plant and define a variable for the objective function
        model = self._plant.to_pyomo(mutable=False)
        objective = 0.0
        # for each node with a price (i.e., suppliers and purchasers) add the cost to the objective function
        if self._cost_weight is not None:
            for cmp in [*self._plant.suppliers, *self._plant.purchasers, *self._plant.machines]:
                objective += self._cost_weight * model.component(cmp).cost
        # for each storage, add the cost for storage difference to the objective function
        if self._storage_weight is not None:
            for storage, weight in self._storage_weight.items():
                cmp = model.component(storage)
                # model a variable for the storage difference (i.e., | storage - current_storage |)
                #  - the absolute value is relaxed to storage_diff >= | storage - current_storage |
                #  - this relaxation allows to linearize the constraint into two different ones
                #  - eventually, the value is multiplied by the respective weight and added to the objective
                cmp.storage_diff = pyo.Var(domain=pyo.NonNegativeReals, initialize=0.0)
                cmp.storage_diff_geq = pyo.Constraint(rule=cmp.storage_diff >= cmp.storage - cmp.current_storage)
                cmp.storage_diff_leq = pyo.Constraint(rule=cmp.storage_diff >= cmp.current_storage - cmp.storage)
                objective += weight * cmp.storage_diff
        # for each machine, add the costs for on/off/state to the objective function
        if self._machine_weight is not None:
            for machine, (on_weight, off_weight, state_weight) in self._machine_weight.items():
                cmp = model.component(machine)
                mac = self._plant.machines[machine]
                # define variables for switch change (use variables instead of expressions to avoid pyomo errors)
                #  - on_diff == 1 if the machine is on (cmp.switch == 1) and it was set as off (was_on == 0)
                #  - off_diff == 1 if the machine is off (cmp.switch == 0) and it was set as on (was_on == 1)
                was_on = 0 if np.isnan(pyo.value(mac.current_state)) else 1
                cmp.on_diff = pyo.Var(domain=pyo.Binary, initialize=0)
                cmp.on_diff_cst = pyo.Constraint(rule=cmp.on_diff == (1 - was_on) * cmp.switch)
                cmp.off_diff = pyo.Var(domain=pyo.Binary, initialize=0)
                cmp.off_diff_cst = pyo.Constraint(rule=cmp.off_diff == was_on * (1 - cmp.switch))
                # define a variable for state change, i.e. state_diff == | node.current_state - node.state |
                #  - this value has a meaning only if the machine was set as on, and it is still on
                #  - otherwise, the on_diff and off_diff variables are considered in the objective
                # in order to model the state difference when the machine is off we rely on the big-M formulation
                #  - M is computed as the maximal setpoint difference, i.e., the largest state
                #  - additionally, we define a binary control variable z = was_on * switch
                #  - using z, we force state_diff <= 0 whenever z == 0, i.e., either the machine was set or is off
                #  - then, we add the constraint state_diff >= | current_state - state | with the term -z * gap
                #  - this last term is used to guarantee that the rhs is negative (i.e, trivial) whenever z == 0
                z = was_on * cmp.switch
                m = mac.setpoint.index[-1]
                cmp.state_diff = pyo.Var(domain=pyo.NonNegativeReals, bounds=(0, m), initialize=0.0)
                cmp.state_diff_m = pyo.Constraint(rule=cmp.state_diff <= z * m)
                cmp.state_diff_geq = pyo.Constraint(rule=cmp.state_diff >= cmp.state - cmp.current_state - (1 - z) * m)
                cmp.state_diff_leq = pyo.Constraint(rule=cmp.state_diff >= cmp.current_state - cmp.state - (1 - z) * m)
                # eventually, multiply each value by the respective weight and add them to the objective
                objective += on_weight * cmp.on_diff + off_weight * cmp.off_diff + state_weight * cmp.state_diff
        # add the objective function to the model, solve it using the defined solver, and eventually return the plan
        model.objective = pyo.Objective(expr=objective, sense=pyo.minimize)
        solver = pyo.SolverFactory(self._solver)
        for option, value in self._solver_options.items():
            solver.options[option] = value
        solver.solve(model)
        plan = {}
        for edge in self._plant.edges().values():
            cmp = model.component(edge.name)
            plan[edge.key] = self._get_value(cmp.flow)
        for machine in self._plant.machines.values():
            cmp = model.component(machine.name)
            plan[machine.key] = self._get_value(cmp.state) if self._get_value(cmp.switch) == 1 else np.nan
        return plan

    def _get_value(self, variable) -> float:
        return np.round(pyo.value(variable), decimals=self._decimals)



from typing import Iterable, Callable, Any, Optional


def get_filtering_function(user_input: Optional) -> Callable[[Any], bool]:
    """Builds a function f(value) -> bool which says whether that value matches or not the user input.

    :param user_input:
        If None is passed, the function always evaluates to true.
        If an iterable object is passed, checks whether the value is in the iterable.
        Otherwise, checks whether the value is exactly the user input passed.

    :return:
        The filtering function f(value) -> bool.
    """
    if user_input is None:
        return lambda d: True
    elif isinstance(user_input, Iterable) and not isinstance(user_input, str):
        user_input = set(user_input)
        return lambda d: d in user_input
    else:
        return lambda d: d == user_input


def get_indexed_object(keys: Optional, indexed: dict) -> Optional[list]:
    """Indexes a set of keys in an indexed dictionary.

    :param keys:
        If None is passed, returns None.
        If a single value is passed, returns a list with the single element <indexed[keys]>.
        If an iterable is passed, returns a list with multiple elements <indexed[k]>, with k in keys.

    :param indexed:
        The dictionary to be indexed.

    :return:
        The list of indexed objects, or None in case the keys are not passed.
    """
    if keys is None:
        return None
    keys = [keys] if isinstance(keys, Iterable) and not isinstance(keys, str) else keys
    return [indexed[k] for k in keys]


def get_matching_object(matcher: Optional, index: Any, default: Any):
    """Uses a matching strategy to return a matching object.

    :param matcher:
        If None is passed, returns the default object.
        If a dictionary is passed, returns the object that matches with the given index.
        Otherwise, returns the matcher itself, since it is the same for every index.

    :param index:
        The matching index in case a dictionary matcher is passed.

    :param default:
        The default object in case a None matcher is passed.

    :return:
        The matching object.
    """
    if matcher is None:
        return default
    elif isinstance(matcher, dict):
        matcher = matcher.get(index)
        return default if matcher is None else matcher
    else:
        return matcher



# Run this script from the origin folder as:
#   > "python setup.py clean" in order to clean previous builds
#   > "python setup.py test" in order to execute all the unittests
#   > "python setup.py sdist" in order to build the library
#
# The package can then be published with:
#   > twine upload dist/*

from setuptools import find_packages, setup

# set up the library metadata and make the build
with open('README.md', 'r') as readme:
    setup(
        name='powerplantsim',
        version='0.1.0',
        maintainer='Luca Giuliani',
        maintainer_email='luca.giuliani13@unibo.it',
        author='University of Bologna - DISI',
        description='PowerPlantSim: a power plant simulator',
        long_description=readme.read(),
        long_description_content_type='text/markdown',
        packages=find_packages(include=['powerplantsim*']),
        python_requires='~=3.10',
        install_requires=[
            'matplotlib~=3.8.2',
            'networkx~=3.2.1',
            'numpy~=1.26.2',
            'pandas~=2.1.3',
            'pyomo~=6.7.0',
            'scikit-learn~=1.3.2',
            'tqdm~=4.66.1'
        ],
        test_suite='test'
    )



import copy
from abc import abstractmethod, ABC
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
import pyomo.environ as pyo
# noinspection PyPackageRequirements
from descriptors import classproperty

from powerplantsim import utils
from powerplantsim.utils.typing import Flow, State


@dataclass(frozen=True, repr=False, eq=False, unsafe_hash=False, kw_only=True)
class DataType(ABC):
    """Abstract class that defines a datatype and has a unique key for comparison."""

    _plant: Any = field(kw_only=True)
    """The power plant object to which this datatype is attached."""

    _info: Dict[str, Any] = field(init=False, default_factory=dict)
    """Internal object for additional mutable information."""

    @classproperty
    def eps(self) -> float:
        """The tolerance to account for numerical errors."""
        return 1e-5

    @property
    @abstractmethod
    def key(self) -> Any:
        """An identifier of the datatype."""
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the datatype."""
        pass

    @classproperty
    def _properties(self) -> List[str]:
        """The list of public properties of the datatype."""
        return ['name']

    @property
    def _step(self) -> Optional[int]:
        """The current step of the simulation, or None if the simulation is not started yet."""
        return self._plant.step

    @property
    def _index(self) -> Optional:
        """The current index of the simulation as in the time horizon, or None if the simulation is not started."""
        return self._plant.index

    @property
    def _horizon(self) -> pd.Index:
        """The time horizon of the simulation in which the datatype is involved."""
        return self._plant.horizon

    @property
    def dict(self) -> Dict[str, Any]:
        """A dictionary containing all the information of the datatype object indexed via property name."""
        return {param: getattr(self, param) for param in self._properties}

    def _instance(self, other) -> bool:
        """Checks whether a different object is matching the self instance for comparison."""
        return isinstance(other, self.__class__)

    def copy(self):
        """Copies the datatype.

        :return:
            A copy of the datatype.
        """
        return copy.deepcopy(self)

    def to_json(self) -> Dict[str, Any]:
        """Function to make the object serializable.

        :return:
            A dictionary containing all the information of the datatype which can be dumped in a json file.
        """
        json = {}
        for param in self._properties:
            value = getattr(self, param)
            if isinstance(value, set):
                value = list(value)
            elif isinstance(value, pd.Series):
                value = value.to_dict()
            elif isinstance(value, pd.DataFrame):
                value = value.to_dict(orient='tight')
            json[param] = value
        return json

    @abstractmethod
    def to_pyomo(self, mutable: bool = False) -> pyo.Block:
        """Encodes the node as a Pyomo block.

        :param mutable:
            Whether to set simulation-specific parameters (e.g., current_price/current_demand/etc.) as mutable and
            without an initialization value, or to initialize them based on their specific value of the current step.

        :return:
            A pyomo.environment.Block object modeling the node.
        """
        pass

    @abstractmethod
    def update(self, rng: np.random.Generator, flows: Dict[Any, Flow], states: Dict[Any, State]):
        """Updates the current internal values of the datatype before the recourse action is called.

        :param rng:
            The random number generator to be used for reproducible results.

        :param flows:
            The user-defined edge flows for the current step, indexed by Edge object.

        :param states:
            The user-defined machine states for the current step, indexed by Machine object.
        """
        pass

    @abstractmethod
    def step(self, flows: Dict[Any, Flow], states: Dict[Any, State]):
        """Performs a step forward in the simulation by computing internal values after the recourse action is called.

        :param flows:
            The edge flows computed by the recourse action for the current step, indexed by Edge object.

        :param states:
            The machine states computed by the recourse action for the current step, indexed by Machine object.
        """
        pass

    def __eq__(self, other: Any) -> bool:
        return self._instance(other) and self.key == other.key

    def __hash__(self) -> int:
        return hash(self.key)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({utils.stringify(value=self.key)})"
